@@ -1,10 +1,10 @@
 from .utils import logger
 from .tree import Tree
-from .visitors import Transformer, Visitor, v_args, Discard
+from .visitors import Transformer, Visitor, v_args, Discard, Transformer_NonRecursive
 from .visitors import InlineTransformer, inline_args   # XXX Deprecated
 from .exceptions import (ParseError, LexError, GrammarError, UnexpectedToken,
                          UnexpectedInput, UnexpectedCharacters, LarkError)
 from .lexer import Token
 from .lark import Lark
 
-__version__ = "0.10.0"
+__version__ = "0.11.1"
