@@ -8,6 +8,7 @@ from .lexer import Token
 ###{standalone
 from inspect import getmembers, getmro
 
+
 class Discard(Exception):
     """When raising the Discard exception in a transformer callback,
     that node is discarded and won't appear in the parent.
@@ -16,6 +17,7 @@ class Discard(Exception):
 
 # Transformers
 
+
 class _Decoratable:
     "Provides support for decorating methods with @v_args"
 
@@ -47,7 +49,7 @@ class _Decoratable:
 class Transformer(_Decoratable):
     """Transformers visit each node of the tree, and run the appropriate method on it according to the node's data.
 
-    Calls its methods (provided by user via inheritance) according to ``tree.data``.
+    Calls its methods (provided by the user via inheritance) according to ``tree.data``.
     The returned value replaces the old one in the structure.
 
     They work bottom-up (or depth-first), starting with the leaves and ending at the root of the tree.
@@ -64,12 +66,11 @@ class Transformer(_Decoratable):
     - ``Transformer_InPlaceRecursive`` - Recursive. Changes the tree in-place instead of returning new instances
 
     Parameters:
-        visit_tokens: By default, transformers only visit rules.
-            visit_tokens=True will tell ``Transformer`` to visit tokens
-            as well. This is a slightly slower alternative to lexer_callbacks
-            but it's easier to maintain and works for all algorithms
-            (even when there isn't a lexer).
+        visit_tokens (bool, optional): Should the transformer visit tokens in addition to rules.
+                                       Setting this to ``False`` is slightly faster. Defaults to ``True``.
+                                       (For processing ignored tokens, use the ``lexer_callbacks`` options)
 
+    NOTE: A transformer without methods essentially performs a non-memoized deepcopy.
     """
     __visit_tokens__ = True   # For backwards compatibility
 
@@ -108,7 +109,6 @@ class Transformer(_Decoratable):
             except Exception as e:
                 raise VisitError(token.type, token, e)
 
-
     def _transform_children(self, children):
         for c in children:
             try:
@@ -126,29 +126,29 @@ class Transformer(_Decoratable):
         return self._call_userfunc(tree, children)
 
     def transform(self, tree):
+        "Transform the given tree, and return the final result"
         return self._transform_tree(tree)
 
     def __mul__(self, other):
+        """Chain two transformers together, returning a new transformer.
+        """
         return TransformerChain(self, other)
 
     def __default__(self, data, children, meta):
-        """Default operation on tree (for override)
+        """Default function that is called if there is no attribute matching ``data``
 
-        Function that is called on if a function with a corresponding name has not been found.
-        Defaults to reconstruct the Tree.
+        Can be overridden. Defaults to creating a new copy of the tree node (i.e. ``return Tree(data, children, meta)``)
         """
         return Tree(data, children, meta)
 
     def __default_token__(self, token):
-        """Default operation on token (for override)
+        """Default function that is called if there is no attribute matching ``token.type``
 
-        Function that is called on if a function with a corresponding name has not been found.
-        Defaults to just return the argument.
+        Can be overridden. Defaults to returning the token as-is.
         """
         return token
 
 
-
 class InlineTransformer(Transformer):   # XXX Deprecated
     def _call_userfunc(self, tree, new_children=None):
         # Assumes tree is already transformed
@@ -175,7 +175,10 @@ class TransformerChain(object):
 
 
 class Transformer_InPlace(Transformer):
-    "Non-recursive. Changes the tree in-place instead of returning new instances"
+    """Same as Transformer, but non-recursive, and changes the tree in-place instead of returning new instances
+
+    Useful for huge trees. Conservative in memory.
+    """
     def _transform_tree(self, tree):           # Cancel recursion
         return self._call_userfunc(tree)
 
@@ -187,7 +190,12 @@ class Transformer_InPlace(Transformer):
 
 
 class Transformer_NonRecursive(Transformer):
-    "Non-recursive. Doesn't change the original tree."
+    """Same as Transformer but non-recursive.
+
+    Like Transformer, it doesn't change the original tree.
+
+    Useful for huge trees.
+    """
 
     def transform(self, tree):
         # Tree to postfix
@@ -195,7 +203,7 @@ class Transformer_NonRecursive(Transformer):
         q = [tree]
         while q:
             t = q.pop()
-            rev_postfix.append( t )
+            rev_postfix.append(t)
             if isinstance(t, Tree):
                 q += t.children
 
@@ -217,15 +225,13 @@ class Transformer_NonRecursive(Transformer):
         return t
 
 
-
 class Transformer_InPlaceRecursive(Transformer):
-    "Recursive. Changes the tree in-place instead of returning new instances"
+    "Same as Transformer, recursive, but changes the tree in-place instead of returning new instances"
     def _transform_tree(self, tree):
         tree.children = list(self._transform_children(tree.children))
         return self._call_userfunc(tree)
 
 
-
 # Visitors
 
 class VisitorBase:
@@ -233,7 +239,10 @@ class VisitorBase:
         return getattr(self, tree.data, self.__default__)(tree)
 
     def __default__(self, tree):
-        "Default operation on tree (for override)"
+        """Default function that is called if there is no attribute matching ``tree.data``
+
+        Can be overridden. Defaults to doing nothing.
+        """
         return tree
 
     def __class_getitem__(cls, _):
@@ -241,18 +250,19 @@ class VisitorBase:
 
 
 class Visitor(VisitorBase):
-    """Bottom-up visitor, non-recursive.
+    """Tree visitor, non-recursive (can handle huge trees).
 
-    Visits the tree, starting with the leaves and finally the root (bottom-up)
-    Calls its methods (provided by user via inheritance) according to ``tree.data``
+    Visiting a node calls its methods (provided by the user via inheritance) according to ``tree.data``
     """
 
     def visit(self, tree):
+        "Visits the tree, starting with the leaves and finally the root (bottom-up)"
         for subtree in tree.iter_subtrees():
             self._call_userfunc(subtree)
         return tree
 
     def visit_topdown(self,tree):
+        "Visit the tree, starting at the root, and ending at the leaves (top-down)"
         for subtree in tree.iter_subtrees_topdown():
             self._call_userfunc(subtree)
         return tree
@@ -261,11 +271,13 @@ class Visitor(VisitorBase):
 class Visitor_Recursive(VisitorBase):
     """Bottom-up visitor, recursive.
 
-    Visits the tree, starting with the leaves and finally the root (bottom-up)
-    Calls its methods (provided by user via inheritance) according to ``tree.data``
+    Visiting a node calls its methods (provided by the user via inheritance) according to ``tree.data``
+
+    Slightly faster than the non-recursive version.
     """
 
     def visit(self, tree):
+        "Visits the tree, starting with the leaves and finally the root (bottom-up)"
         for child in tree.children:
             if isinstance(child, Tree):
                 self.visit(child)
@@ -274,6 +286,7 @@ class Visitor_Recursive(VisitorBase):
         return tree
 
     def visit_topdown(self,tree):
+        "Visit the tree, starting at the root, and ending at the leaves (top-down)"
         self._call_userfunc(tree)
 
         for child in tree.children:
@@ -283,7 +296,6 @@ class Visitor_Recursive(VisitorBase):
         return tree
 
 
-
 def visit_children_decor(func):
     "See Interpreter"
     @wraps(func)
@@ -324,8 +336,6 @@ class Interpreter(_Decoratable):
         return self.visit_children(tree)
 
 
-
-
 # Decorators
 
 def _apply_decorator(obj, decorator, **kwargs):
@@ -337,7 +347,6 @@ def _apply_decorator(obj, decorator, **kwargs):
         return _apply(decorator, **kwargs)
 
 
-
 def _inline_args__func(func):
     @wraps(func)
     def create_decorator(_f, with_self):
@@ -356,7 +365,6 @@ def inline_args(obj):   # XXX Deprecated
     return _apply_decorator(obj, _inline_args__func)
 
 
-
 def _visitor_args_func_dec(func, visit_wrapper=None, static=False):
     def create_decorator(_f, with_self):
         if with_self:
@@ -376,11 +384,11 @@ def _visitor_args_func_dec(func, visit_wrapper=None, static=False):
     return f
 
 
-def _vargs_inline(f, data, children, meta):
+def _vargs_inline(f, _data, children, _meta):
     return f(*children)
-def _vargs_meta_inline(f, data, children, meta):
+def _vargs_meta_inline(f, _data, children, meta):
     return f(meta, *children)
-def _vargs_meta(f, data, children, meta):
+def _vargs_meta(f, _data, children, meta):
     return f(children, meta)   # TODO swap these for consistency? Backwards incompatible!
 def _vargs_tree(f, data, children, meta):
     return f(Tree(data, children, meta))
@@ -394,10 +402,14 @@ def v_args(inline=False, meta=False, tree=False, wrapper=None):
     ``v_args`` can modify this behavior. When used on a transformer/visitor class definition,
     it applies to all the callback methods inside it.
 
+    ``v_args`` can be applied to a single method, or to an entire class. When applied to both,
+    the options given to the method take precedence.
+
     Parameters:
-        inline: Children are provided as ``*args`` instead of a list argument (not recommended for very long lists).
-        meta: Provides two arguments: ``children`` and ``meta`` (instead of just the first)
-        tree: Provides the entire tree as the argument, instead of the children.
+        inline (bool, optional): Children are provided as ``*args`` instead of a list argument (not recommended for very long lists).
+        meta (bool, optional): Provides two arguments: ``children`` and ``meta`` (instead of just the first)
+        tree (bool, optional): Provides the entire tree as the argument, instead of the children.
+        wrapper (function, optional): Provide a function to decorate all methods.
 
     Example:
         ::
@@ -440,7 +452,7 @@ def v_args(inline=False, meta=False, tree=False, wrapper=None):
 ###}
 
 
-#--- Visitor Utilities ---
+# --- Visitor Utilities ---
 
 class CollapseAmbiguities(Transformer):
     """
@@ -454,7 +466,9 @@ class CollapseAmbiguities(Transformer):
     """
     def _ambig(self, options):
         return sum(options, [])
+
     def __default__(self, data, children_lists, meta):
         return [Tree(data, children, meta) for children in combine_alternatives(children_lists)]
+
     def __default_token__(self, t):
         return [t]
